#
#                _         _    _      _                _    
#               (_)       | |  | |    | |              | |   
#  _ __    ___   _  _ __  | |_ | |__  | |  __ _  _ __  | | __
# | '_ \  / _ \ | || '_ \ | __|| '_ \ | | / _` || '_ \ | |/ /
# | |_) || (_) || || | | || |_ | |_) || || (_| || | | ||   < 
# | .__/  \___/ |_||_| |_| \__||_.__/ |_| \__,_||_| |_||_|\_\
# | |                                                        
# |_|                                                        
# 
# This file is part of the 'rich-iannone/pointblank' package.
# 
# (c) Richard Iannone <riannone@me.com>
# 
# For full copyright and license information, please look at
# https://rich-iannone.github.io/pointblank/LICENSE.html
#


#' Get a summary report from an agent
#' 
#' @description 
#' We can get an informative summary table from an agent by using the
#' `get_agent_report()` function. The table can be provided in two substantially
#' different forms: as a **gt** based display table (the default), or, as a
#' tibble. The amount of fields with intel is different depending on whether or
#' not the agent performed an interrogation (with the [interrogate()] function).
#' Basically, before [interrogate()] is called, the agent will contain just the
#' validation plan (however many rows it has depends on how many validation
#' functions were supplied a part of that plan). Post-interrogation, information
#' on the passing and failing test units is provided, along with indicators on
#' whether certain failure states were entered (provided they were set through
#' `actions`). The display table variant of the agent report, the default form,
#' will have the following columns:
#' 
#' \itemize{
#' \item i (unlabeled): the validation step number
#' \item STEP: the name of the validation function used for the validation step
#' \item COLUMNS: the names of the target columns used in the validation step
#' (if applicable)
#' \item VALUES: the values used in the validation step, where applicable; this
#' could be as literal values, as column names, an expression, a set of
#' sub-validations (for a [conjointly()] validation step), etc.
#' \item TBL: indicates whether any there were any preconditions to apply
#' before interrogation; if not, a script 'I' stands for 'identity' but, if so,
#' a right-facing arrow appears
#' \item EVAL: a character value that denotes the result of each validation
#' step functions' evaluation during interrogation
#' \item *N*: the total number of test units for the validation step
#' \item PASS: the number of test units that received a *pass*
#' \item FAIL: the fraction of test units that received a *pass*
#' \item W, S, N: indicators that show whether the `warn`, `stop`, or `notify`
#' states were entered; unset states appear as dashes, states that are set with
#' thresholds appear as unfilled circles when not entered and filled when
#' thresholds are exceeded (colors for W, S, and N are amber, red, and blue)
#' \item EXT: a column that provides buttons with data extracts for each
#' validation step where failed rows are available (as CSV files)
#' }
#' 
#' The small version of the display table (obtained using `size = "small"`)
#' omits the `COLUMNS`, `TBL`, and `EXT` columns. The width of the small table
#' is 575px; the standard table is 875px wide.
#' 
#' If choosing to get a tibble (with `display_table = FALSE`), it will have the
#' following columns:
#' 
#' \itemize{
#' \item i: the validation step number
#' \item type: the name of the validation function used for the validation step
#' \item columns: the names of the target columns used in the validation step
#' (if applicable)
#' \item values: the values used in the validation step, where applicable; for
#' a [conjointly()] validation step, this is a listing of all sub-validations
#' \item precon: indicates whether any there are any preconditions to apply
#' before interrogation and, if so, the number of statements used
#' \item active: a logical value that indicates whether a validation step is
#' set to `"active"` during an interrogation
#' \item eval: a character value that denotes the result of each validation
#' step functions' evaluation during interrogation
#' \item units: the total number of test units for the validation step
#' \item n_pass: the number of test units that received a *pass*
#' \item f_pass: the fraction of test units that received a *pass*
#' \item W, S, N: logical value stating whether the `warn`, `stop`, or `notify`
#' states were entered
#' \item extract: a logical value that indicates whether a data extract is
#' available for the validation step
#' }
#' 
#' @param agent An agent object of class `ptblank_agent`.
#' @param arrange_by A choice to arrange the report table rows by the validation
#'   step number (`"i"`, the default), or, to arrange in descending order by
#'   severity of the failure state (with `"severity"`).
#' @param keep An option to keep `"all"` of the report's table rows (the
#'   default), or, keep only those rows that reflect one or more
#'   `"fail_states"`.
#' @param display_table Should a display table be generated? If `TRUE` (the
#'   default), and if the **gt** package is installed, a display table for the
#'   report will be shown in the Viewer. If `FALSE`, or if **gt** is not
#'   available, then a tibble will be returned.
#' @param size The size of the display table, which can be either `"standard"`
#'   (the default) or `"small"`. This only applies to a display table (where
#'   `display_table = TRUE`).
#' @param title Options for customizing the title of the report. The default is
#'   the keyword `":default:"` which produces generic title text that refers to
#'   the **pointblank** package in the language governed by the `lang` option.
#'   Another keyword option is `":tbl_name:"`, and that presents the name of the
#'   table as the title for the report. If no title is wanted, then the
#'   `":none:"` keyword option can be used. Aside from keyword options, text can
#'   be provided for the title and `glue::glue()` calls can be used to construct
#'   the text string. If providing text, it will be interpreted as Markdown text
#'   and transformed internally to HTML. To circumvent such a transformation,
#'   use text in [I()] to explicitly state that the supplied text should not be
#'   transformed.
#' @param lang The language to use for automatic creation of briefs (short
#'   descriptions for each validation step) and for the *agent report* (a
#'   summary table that provides the validation plan and the results from the
#'   interrogation. By default, `NULL` will create English (`"en"`) text. Other
#'   options include French (`"fr"`), German (`"de"`), Italian (`"it"`), Spanish
#'   (`"es"`), Portuguese (`"pt"`), Turkish (`"tr"`), Chinese (`"zh"`), Russian
#'   (`"ru"`), Polish (`"pl"`), Danish (`"da"`), Swedish (`"sv"`), and Dutch
#'   (`"nl"`). This `lang` option will override any previously set language
#'   setting (e.g., by the [create_agent()] call).
#' @param locale An optional locale ID to use for formatting values in the
#'   *agent report* summary table according the locale's rules. Examples include
#'   `"en_US"` for English (United States) and `"fr_FR"` for French (France);
#'   more simply, this can be a language identifier without a country
#'   designation, like `"es"` for Spanish (Spain, same as `"es_ES"`). This
#'   `locale` option will override any previously set locale value (e.g., by the
#'   [create_agent()] call).
#' 
#' @return A **gt** table object if `display_table = TRUE` or a tibble if
#'   `display_table = FALSE`.
#' 
#' @examples
#' # Create a simple table with a
#' # column of numerical values
#' tbl <- 
#'   dplyr::tibble(a = c(5, 7, 8, 5))
#' 
#' # Validate that values in column
#' # `a` are always greater than 4
#' agent <-
#'   create_agent(tbl = tbl) %>%
#'   col_vals_gt(vars(a), value = 4) %>%
#'   interrogate()
#' 
#' # Get a tibble-based report from the
#' # agent by using `get_agent_report()`
#' # with `display_table = FALSE`
#' agent %>%
#'   get_agent_report(display_table = FALSE)
#'   
#' # View a the report by printing the
#' # `agent` object anytime, but, return a
#' # gt table object by using this with
#' # `display_table = TRUE` (the default)
#' report <- get_agent_report(agent)
#' class(report)
#' 
#' # What can you do with the report?
#' # Print it from an R Markdown code,
#' # use it in an email, put it in a
#' # webpage, or further modify it with
#' # the **gt** package
#' 
#' # The agent report as a **gt** display
#' # table comes in two sizes: "standard"
#' # (the default) and "small"
#' small_report <- 
#'   get_agent_report(
#'     agent = agent,
#'     size = "small"
#'   )
#' 
#' class(small_report)
#' 
#' # The standard report is 875px wide
#' # the small one is 575px wide
#' 
#' @family Interrogate and Report
#' @section Function ID:
#' 6-2
#' 
#' @export
get_agent_report <- function(agent,
                             arrange_by = c("i", "severity"),
                             keep = c("all", "fail_states"),
                             display_table = TRUE,
                             size = "standard",
                             title = ":default:",
                             lang = NULL,
                             locale = NULL) {
  
  arrange_by <- match.arg(arrange_by)
  keep <- match.arg(keep)
  
  validation_set <- agent$validation_set
  
  eval <- 
    validation_set %>%
    dplyr::select(eval_error, eval_warning) %>%
    dplyr::mutate(condition = dplyr::case_when(
      !eval_error & !eval_warning ~ "OK",
      eval_error & eval_warning ~ "W + E",
      eval_error ~ "ERROR",
      eval_warning ~ "WARNING"
    )) %>%
    dplyr::pull(condition)
  
  columns <- 
    validation_set$column %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        ifelse(
          is.null(x),
          NA_character_,
          unlist(x)
        )
      }
    )
  
  values <- 
    validation_set$values %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        ifelse(
          is.null(x),
          NA_character_,
          paste(x %>% gsub("~", "", .), collapse = ", ")
        )
      } 
    )
  
  precon_count <-
    validation_set$preconditions %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        ifelse(
          is.null(x),
          NA_character_,
          x %>% rlang::as_function() %>% length() %>% as.character()
        )
      }
    )
  
  if (!has_agent_intel(agent)) {
    
    extract_count <- rep(NA, nrow(validation_set))
    
  } else {
    
    extract_count <- 
      as.character(validation_set[["i"]]) %in% names(agent$extracts)
    
    extract_count[extract_count == FALSE] <- NA_integer_
    
    suppressWarnings(
      extract_count[!is.na(extract_count)] <- 
        vapply(
          agent$extracts,
          FUN.VALUE = integer(1),
          USE.NAMES = FALSE,
          FUN = nrow
        )
    )
  }
  
  report_tbl <- 
    dplyr::tibble(
      i = validation_set$i,
      type = validation_set$assertion_type,
      columns = columns,
      values = values,
      precon = precon_count,
      active = validation_set$eval_active,
      eval = eval,
      units = validation_set$n,
      n_pass = validation_set$n_passed,
      f_pass = validation_set$f_passed,
      W = validation_set$warn,
      S = validation_set$stop,
      N = validation_set$notify,
      extract = extract_count
    )
  
  report_tbl <-
    report_tbl %>%
    dplyr::mutate(
      eval_pts = ifelse(eval != "OK", 10, 0),
      N_pts = ifelse(!is.na(N) & N, 3, 0),
      S_pts = ifelse(!is.na(S) & S, 2, 0),
      W_pts = ifelse(!is.na(W) & W, 1, 0),
      total_pts = eval_pts + N_pts + S_pts + W_pts
    )
  
  if (arrange_by == "severity") {
    report_tbl <-
      report_tbl %>%
      dplyr::arrange(dplyr::desc(total_pts))
  }
  
  if (keep == "fail_states") {
    report_tbl <- report_tbl %>% dplyr::filter(total_pts > 0)
  }
  
  report_tbl <-
    report_tbl %>%
    dplyr::select(-dplyr::ends_with("pts"))
  
  if (!display_table) {
    return(report_tbl)
  }
  
  # Generate a gt table if `display_table == TRUE`
  
  # nocov start
  validation_set <- validation_set[report_tbl$i, ]
  eval <- eval[report_tbl$i]
  extracts <- 
    agent$extracts[
      as.character(
        base::intersect(as.numeric(names(agent$extracts)), report_tbl$i)
      )
    ]
  
  if (is.null(lang)) {
    
    lang <- agent$lang
    if (is.null(locale)) locale <- agent$locale
    
  } else {
    
    normalize_reporting_language(lang = lang)
    
    # Set the `locale` to the `lang` value if `locale` isn't set
    if (is.null(locale)) locale <- lang
  }

  briefs <- validation_set$brief
  assertion_type <- validation_set$assertion_type
  label <- validation_set$label
  tbl_src <- agent$tbl_src
  tbl_name <- agent$tbl_name
  
  # Generate agent label HTML
  agent_label_styled <- create_agent_label_html(agent)
  
  # Generate table type HTML
  table_type <- 
    create_table_type_html(
      tbl_src = tbl_src,
      tbl_name = tbl_name
    )

  # Generate action levels HTML
  action_levels <- make_action_levels_html(agent, locale)
  
  # Combine label, table type, and action levels into
  # a table subtitle <div>
  combined_subtitle <-
    htmltools::tagList(
      htmltools::HTML(agent_label_styled),
      htmltools::tags$div(
        style = htmltools::css(
          height = "25px"
        ),
        htmltools::HTML(paste0(table_type, action_levels))
      ) 
    ) %>% as.character()

  # Generate table execution start/end time (and duration)
  # as a table source note
  table_time <- 
    create_table_time_html(
      time_start = agent$time_start,
      time_end = agent$time_end,
      size = size,
      locale = locale
    )
  
  # Reformat `type`
  type_upd <- 
    seq_along(assertion_type) %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        
        assertion_type_nchar <- nchar(assertion_type[x])
        
        text_size <- ifelse(assertion_type_nchar + 2 >= 20, 10, 11)
        text_size <- ifelse(size == "small", 11, text_size)
        
        if (size == "small") {

          htmltools::tags$code(
            style = htmltools::css(
              `font-size` = paste0(text_size, "px")
            ),
            htmltools::HTML(paste0("&nbsp;", assertion_type[x], "()"))
          ) %>%
            as.character()
          
        } else {
          
          if (!is.na(label[x])) {

            htmltools::tags$div(
              `aria-label` = briefs[x],
              `data-balloon-pos` = "right",
              style = htmltools::css(width = "fit-content"),
              htmltools::tags$div(
                style = htmltools::css(
                  float = "left",
                  width = "30px"
                ),
                htmltools::HTML(add_icon_svg(icon = assertion_type[x]))
              ),
              htmltools::tags$div(
                style = htmltools::css(
                  `line-height` = "0.7em",
                  `margin-left` = "32px",
                  `padding-left` = "3px"
                ),
                htmltools::tags$p(
                  style = htmltools::css(
                    margin = "0",
                    `padding-top` = "4px",
                    `font-size` = "9px"
                  ),
                  htmltools::HTML(label[x])
                ),
                htmltools::tags$p(
                  style = htmltools::css(margin = "0"),
                  htmltools::tags$code(
                    style = htmltools::css(`font-size` = "11px"),
                    paste0(assertion_type[x], "()")
                  )
                )
              )
            ) %>% 
              as.character()

          } else {

            htmltools::tags$div(
              `aria-label` = briefs[x],
              `data-balloon-pos` = "right",
              style = htmltools::css(width = "fit-content"),
              htmltools::HTML(add_icon_svg(icon = assertion_type[x])),
              htmltools::tags$code(
                style = htmltools::css(`font-size` = paste0(text_size, "px")),
                htmltools::HTML(paste0("&nbsp;", assertion_type[x], "()"))
              )
            ) %>%
              as.character()
          }
        }
      }
    )
  
  # Reformat `columns`
  columns_upd <- 
    validation_set$column %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        
        if (is.null(x) | (is.list(x) && is.na(unlist(x)))) {
          x <- NA_character_
        } else if (is.na(x)) {
          x <- NA_character_
        } else {

          text <- 
            x %>%
            unlist() %>%
            strsplit(", ") %>%
            unlist()
          
          title <- text
          
          text_fragments <- 
            paste0(
              htmltools::tags$span(
                style = htmltools::css(color = "purple"),
                htmltools::HTML("&marker;")
              ), text, collapse = ", "
            )
          
          if (size == "small") {
            
            x <- 
              htmltools::tags$div(
                htmltools::tags$p(
                  title = paste(title, collapse = ", "),
                  style = htmltools::css(
                    `margin-top` = "0",
                    `margin-bottom` = "0",
                    `font-family` = "monospace",
                    `white-space` = "nowrap",
                    `text-overflow` = "ellipsis",
                    overflow = "hidden"
                  ),
                  htmltools::HTML(text_fragments)
                )
              ) %>% 
              as.character()
            
          } else {
            
            x <-
              htmltools::tags$div(
                `aria-label` = paste(title, collapse = ", "),
                `data-balloon-pos` = "left",
                htmltools::tags$p(
                  style = htmltools::css(
                    `margin-top` = "0",
                    `margin-bottom` = "0",
                    `font-size` = "11px",
                    `white-space` = "nowrap",
                    `text-overflow` = "ellipsis",
                    overflow = "hidden",
                    `line-height` = "2em"
                  ),
                  htmltools::tags$code(htmltools::HTML(text_fragments))
                )
              ) %>% 
              as.character()
          }
        }
        x
      }
    )
  
  # Reformat `values`
  values_upd <- 
    validation_set$values %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        
        if (!is.null(x) &&
            !is.null(names(x)) &&
            all(names(x) %in% c("TRUE", "FALSE"))) {

          # Case of in-between comparison validation where there are
          # one or two columns specified as bounds
          bounds_incl <- as.logical(names(x))
          
          if (rlang::is_quosure(x[[1]])) {
            x_left <- 
              paste0(
                "<span style=\"color: purple;\">&marker;</span>",
                rlang::as_label(x[[1]])
              )
          } else {
            x_left <- 
              pb_fmt_number(
                x[[1]],
                decimals = 4,
                drop_trailing_zeros = TRUE,
                locale = locale
              )
          }
          
          if (rlang::is_quosure(x[[2]])) {
            x_right <- 
              paste0(
                "<span style=\"color: purple;\">&marker;</span>",
                rlang::as_label(x[[2]])
              )
          } else {
            x_right <-
              pb_fmt_number(
                x[[2]],
                decimals = 4,
                drop_trailing_zeros = TRUE,
                locale = locale
              )
          }
          
          text <-
            paste0(
              ifelse(bounds_incl[1], "[", "("),
              x_left,
              ", ",
              x_right,
              ifelse(bounds_incl[2], "]", ")")
            )
          
          title <- 
            paste0(
              ifelse(bounds_incl[1], "[", "("),
              pb_fmt_number(
                rlang::as_label(x[[1]]),
                decimals = 4,
                drop_trailing_zeros = TRUE,
                locale = locale
              ),
              ", ",
              pb_fmt_number(
                rlang::as_label(x[[2]]),
                decimals = 4,
                drop_trailing_zeros = TRUE,
                locale = locale
              ),
              ifelse(bounds_incl[2], "]", ")")
            )
          
          if (size == "small") {
            
            x <-
              paste0(
                "<div><p title=\"", title, "\" style=\"margin-top: 0px; ",
                "margin-bottom: 0px; ",
                "font-family: monospace; white-space: nowrap; ",
                "text-overflow: ellipsis; overflow: hidden;\">",
                text,
                "</p></div>"
              )
            
          } else {
            
            x <- 
              paste0(
                "<div aria-label=\"", title, "\" data-balloon-pos=\"left\">",
                "<p style=\"margin-top: 0px; margin-bottom: 0px; ",
                "font-size: 11px; white-space: nowrap; ",
                "text-overflow: ellipsis; overflow: hidden;\">",
                "<code>", text, "</code>",
                "</p></div>"
              )
          }
          
        } else if (is.list(x) && length(x) > 0 && inherits(x, "col_schema")) {
          
          # Case of column schema as a value
          
          column_schema_text <- 
            get_lsv("agent_report/report_column_schema")[[lang]]
          
          column_schema_type_text <- 
            if (inherits(x, "r_type")) {
              get_lsv("agent_report/report_r_col_types")[[lang]]
            } else {
              get_lsv("agent_report/report_r_sql_types")[[lang]]
            }
          
          x <- 
            paste0(
              "<div>",
              "<p style=\"margin-top: 0px; margin-bottom: 0px; ",
              "font-size: 0.75rem;\">", column_schema_text, "</p>",
              "<p style=\"margin-top: 2px; margin-bottom: 0px; ",
              "font-size: 0.65rem;\">", column_schema_type_text, "</p>",
              "</div>"
            )
          
        } else if (is_call(x)) {
          
          text <- rlang::as_label(x)

          if (size == "small") {
            x <- 
              paste0(
                "<div><p title=\"", text, "\" style=\"margin-top: 0px; ",
                "margin-bottom: 0px; ",
                "font-family: monospace; white-space: nowrap; ",
                "text-overflow: ellipsis; overflow: hidden;\">",
                text,
                "</p></div>"
              )
          } else {
            x <- 
              paste0(
                "<div aria-label=\"", text, "\" data-balloon-pos=\"left\">",
                "<p style=\"margin-top: 0px; margin-bottom: 0px; ",
                "font-size: 11px; white-space: nowrap; ",
                "text-overflow: ellipsis; overflow: hidden;\">",
                "<code>", text, "</code>",
                "</p></div>"
              )
          }
          
        } else if (is.list(x) && length(x) > 0 && !inherits(x, "quosures")) {
          
          # Conjointly case
          
          if (length(x) > 1) {
            
            step_text <- 
              paste0(
                length(x), " ",
                get_lsv("agent_report/report_col_steps")[[lang]]
              )
            
          } else {
            
            step_text <- 
              paste0(
                length(x), " ",
                get_lsv("agent_report/report_col_step")[[lang]]
              )
          }
          
          x <- 
            paste0(
              "<div><p style=\"margin-top: 0px; margin-bottom: 0px; ",
              "font-size: 0.75rem;\">", step_text, "</p></div>"
            )
          
        } else if (is.null(x)) {
          
          x <- NA_character_
          
        } else {
          
          text <-
            x %>%
            tidy_gsub(
              "~",
              "<span style=\"color: purple;\">&marker;</span>"
            ) %>%
            unname()
          
          text <- paste(text, collapse = ", ")
          
          if (size == "small") {
            x <- 
              paste0(
                "<div><p title=\"",
                x %>% tidy_gsub("~", "") %>% paste(., collapse = ", "),
                "\" style=\"margin-top: 0px; margin-bottom: 0px; ",
                "font-size: 11px; white-space: nowrap; ",
                "text-overflow: ellipsis; overflow: hidden;\">",
                text,
                "</p></div>"
              )
          } else {
            x <- 
              paste0(
                "<div aria-label=\"",
                x %>% tidy_gsub("~", "") %>% paste(., collapse = ", "),
                "\" data-balloon-pos=\"left\"><p style=\"margin-top: 0px; ",
                "margin-bottom: 0px; ",
                "font-size: 11px; white-space: nowrap; ",
                "text-overflow: ellipsis; overflow: hidden;\">",
                "<code>", text, "</code>",
                "</p></div>"
              )
          }
        }
        x
      } 
    )
  
  # Reformat `precon`
  precon_upd <- 
    validation_set$preconditions %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        
        if (is.null(x)) {
          
          x <- 
            make_boxed_text_html(
              x = icon_status("unchanged"),
              size = size,
              color = "#333333",
              background = "transparent",
              font_size = "10px",
              padding = 0,
              tt_text = get_lsv(text = c(
                "agent_report",
                "report_no_table_preconditions"
              ))[[lang]],
              border_radius = "4px"
            )
          
        } else if (rlang::is_formula(x) || rlang::is_function(x)) {
          
          if (rlang::is_formula(x)) {
            text <- rlang::as_label(x) %>% tidy_gsub("^~", "")
          } else {
            text <- rlang::as_label(body(x))
          }
          
          x <- 
            make_boxed_text_html(
              x = icon_status("modified"),
              size = size,
              color = "#3C898A",
              background = "transparent",
              font_size = "18px",
              padding = 0,
              tt_text = get_lsv(text = c(
                "agent_report",
                "report_some_table_preconditions"
              ))[[lang]],
              border_radius = "4px"
            )
        }
        x
      } 
    )
  
  seg_col <- validation_set$seg_col
  seg_val <- validation_set$seg_val
  
  # Make changes to the `precon` column if there is grouping
  precon_upd <- 
    seq_along(seg_col) %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        
        if (is.na(seg_col[x])) {
          return(precon_upd[x])
        }
        
        seg_col_x <- seg_col[x]
        seg_val_x <- seg_val[x]
        
        precon_upd[x] <- 
          make_boxed_text_html(
            x = icon_status("segmented"),
            size = size, 
            color = "#3C898A",
            background = "transparent",
            font_size = "10px",
            padding = 0,
            tt_text = glue::glue(get_lsv(text = c(
              "agent_report",
              "report_on_segmentation"
            ))[[lang]]),
            border_radius = "4px"
          )
        
        precon_upd[x]
      }
    )
  
  # Reformat `eval`
  eval_upd <- 
    seq_along(eval) %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        
        if (is.na(eval[x])) {
          
          out <- "&mdash;"
          
        } else if (eval[x] == "OK") {
          
          out <- 
            make_boxed_text_html(
              x = "&check;",
              size = size,
              color = "#4CA64C",
              background = "transparent",
              tt_text = get_lsv(text = c(
                "agent_report",
                "report_no_evaluation_issues"
              ))[[lang]],
            )
          
        } else if (eval[x] == "W + E") {
          
          text <- 
            htmltools::htmlEscape(
              agent$validation_set$capture_stack[[x]]$error %>%
                tidy_gsub("\"", "'")
            )
          
          if (!is.null(text)) {
            text <- as.character(text)
            text_size <- "large"
          } else {
            text <- ""
            text_size <- NULL
          }
          
          out <- 
            make_boxed_text_html(
              x = "&#128165;",
              size = size,
              color = "#FFFFFF",
              background = "transparent",
              tt_text = text,
              tt_text_size = text_size
            )
          
        } else if (eval[x] == "WARNING") {
          
          text <- 
            htmltools::htmlEscape(
              agent$validation_set$capture_stack[[x]]$warning %>%
                tidy_gsub("\"", "'")
            )
          
          if (!is.null(text)) {
            text <- as.character(text)
            text_size <- "large"
          } else {
            text <- ""
            text_size <- NULL
          }
          
          out <- 
            make_boxed_text_html(
              x = "&#9888;",
              size = size,
              color = "#222222",
              background = "transparent",
              tt_text = text,
              tt_text_size = text_size
            )
          
        } else if (eval[x] == "ERROR") {
          
          text <-
            htmltools::htmlEscape(
              agent$validation_set$capture_stack[[x]]$error %>%
                tidy_gsub("\"", "'")
            )
          
          if (!is.null(text)) {
            text <- as.character(text)
            text_size <- "large"
          } else {
            text <- ""
            text_size <- NULL
          }
          
          out <- 
            make_boxed_text_html(
              x = "&#128165;",
              size = size,
              color = "#FFFFFF",
              background = "transparent",
              tt_text = text,
              tt_text_size = text_size
            )
        }
        out
      } 
    )
  
  # Reformat `extract`
  extract_upd <-
    validation_set$i %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        
        if (is.null(extracts[as.character(x)][[1]])) {
          x <- "&mdash;"
        } else {
          
          df <- 
            extracts[as.character(x)][[1]] %>%
            as.data.frame(stringsAsFactors = FALSE)

          fail_rows_extract <- 
            pb_fmt_number(nrow(df), decimals = 0, locale = locale)
          
          title_text <- 
            glue::glue(
              get_lsv(
                text = c(
                  "agent_report",
                  "report_fail_rows_available"
                )
              )[[lang]]
            )
          
          temp_file <- 
            tempfile(pattern = paste0("csv_file_", x), fileext = ".csv")
          
          utils::write.csv(df, file = temp_file, row.names = FALSE)
          
          on.exit(unlink(temp_file))
          
          file_encoded <- base64enc::base64encode(temp_file)
          
          output_file_name <- 
            paste0(
              "extract_",
              formatC(x, width = 4, format = "d", flag = "0"),
              ".csv"
            )
          
          x <- 
            htmltools::a(
              href = paste0("data:text/csv;base64,", file_encoded),
              download = output_file_name,
              htmltools::tags$button(
                `aria-label` = title_text,
                `data-balloon-pos` = "left",
                style = htmltools::css(
                  `background-color` = "#67C2DC",
                  color = "#FFFFFF",
                  border = "none",
                  padding = "5px",
                  `font-weight` = "bold",
                  cursor = "pointer",
                  `border-radius` = "4px"
                ),
                "CSV"
              )
            ) %>%
            as.character()
        }
        x
      } 
    )
  
  #
  # Reformat W, S, and N
  #
  
  w_upd <- 
    validation_set$warn %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        if (is.na(x)) {
          x <- "&mdash;"
        } else if (x == TRUE) {
          x <- "<span style=\"color: #E5AB00;\">&#9679;</span>"
        } else if (x == FALSE) {
          x <- "<span style=\"color: #E5AB00;\">&cir;</span>"
        }
        x
      }
    )
  
  s_upd <- 
    validation_set$stop %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        if (is.na(x)) {
          x <- "&mdash;"
        } else if (x == TRUE) {
          x <- "<span style=\"color: #CF142B;\">&#9679;</span>"
        } else if (x == FALSE) {
          x <- "<span style=\"color: #CF142B;\">&cir;</span>"
        }
        x
      }
    )
  
  n_upd <- 
    validation_set$notify %>%
    vapply(
      FUN.VALUE = character(1),
      USE.NAMES = FALSE,
      FUN = function(x) {
        if (is.na(x)) {
          x <- "&mdash;"
        } else if (x == TRUE) {
          x <- "<span style=\"color: #439CFE;\">&#9679;</span>"
        } else if (x == FALSE) {
          x <- "<span style=\"color: #439CFE;\">&cir;</span>"
        }
        x
      }
    )
  
  f_pass_val <- report_tbl$f_pass
  f_pass_val <- ifelse(f_pass_val > 0 & f_pass_val < 0.01, 0.01, f_pass_val)
  f_pass_val <- ifelse(f_pass_val < 1 & f_pass_val > 0.99, 0.99, f_pass_val)
  f_pass_val <- as.numeric(f_pass_val)
  
  f_fail_val <- 1 - report_tbl$f_pass
  f_fail_val <- ifelse(f_fail_val > 0 & f_fail_val < 0.01, 0.01, f_fail_val)
  f_fail_val <- ifelse(f_fail_val < 1 & f_fail_val > 0.99, 0.99, f_fail_val)
  f_fail_val <- as.numeric(f_fail_val)

  
  # Generate the report title with the `title` option
  title_text <- 
    process_title_text(
      title = title,
      tbl_name = tbl_name,
      report_type = "agent",
      lang = lang
    )
  
  # Generate a gt table
  agent_report <- 
    report_tbl %>%
    dplyr::mutate(
      status_color = NA_character_,
      type = type_upd,
      columns = columns_upd,
      values = values_upd,
      precon = precon_upd,
      eval_sym = eval_upd,
      units = units,
      n_pass = n_pass,
      n_fail = units - n_pass,
      f_pass = f_pass_val,
      f_fail = f_fail_val,
      W_val = W,
      S_val = S,
      N_val = N,
      W = w_upd,
      S = s_upd,
      N = n_upd,
      extract = extract_upd
    ) %>%
    dplyr::select(
      status_color, i, type, columns, values, precon, eval_sym, units,
      n_pass, f_pass, n_fail, f_fail, W, S, N, extract,
      W_val, S_val, N_val, eval, active
    ) %>%
    gt::gt(id = "pb_agent") %>%
    gt::tab_header(
      title = title_text,
      subtitle = gt::md(combined_subtitle)
    ) %>%
    gt::cols_merge(
      columns = c("n_pass", "f_pass"),
      hide_columns = "f_pass"
    ) %>%
    gt::cols_merge(
      columns = c("n_fail", "f_fail"),
      hide_columns = "f_fail"
    ) %>%
    gt::text_transform(
      locations = gt::cells_body(columns = c("n_pass", "n_fail")),
      fn = function(x) {
        dplyr::case_when(
          x == "NA NA"  ~ "&mdash;",
          TRUE ~ x %>%
            tidy_gsub(" ", "</code><br><code>") %>%
            paste0("<code>", ., "</code>")
        )
      }
    ) %>%
    gt::cols_label(
      status_color = "",
      i = "",
      type = get_lsv("agent_report/report_col_step")[[lang]],
      columns = get_lsv("agent_report/report_col_columns")[[lang]],
      values = get_lsv("agent_report/report_col_values")[[lang]],
      precon = "TBL",
      eval_sym = "EVAL",
      units = gt::md("&sdot;&nbsp;&sdot;&nbsp;&sdot;"),
      n_pass = "PASS",
      n_fail = "FAIL",
      extract = "EXT"
    ) %>%
    gt::tab_source_note(source_note = gt::md(table_time)) %>%
    gt::tab_options(
      table.font.size = gt::pct(90),
      row.striping.include_table_body = FALSE
    ) %>%
    gt::cols_align(
      align = "center",
      columns = c("precon", "eval_sym", "W", "S", "N", "extract")
    ) %>%
    gt::cols_align(
      align = "center",
      columns = c("f_pass", "f_fail")
    ) %>%
    gt::cols_align(
      align = "right",
      columns = "i"
    ) %>%
    gt::fmt_number(
      columns = c("units", "n_pass", "n_fail", "f_pass", "f_fail"),
      decimals = 0, drop_trailing_zeros = TRUE, suffixing = TRUE,
      locale = locale
    ) %>%
    gt::fmt_number(
      columns = c("f_pass", "f_fail"),
      decimals = 2,
      locale = locale
    ) %>%
    gt::fmt_markdown(
      columns = c(
        "type", "columns", "values", "precon",
        "eval_sym", "W", "S", "N", "extract"
      )
    ) %>%
    gt::fmt_missing(columns = c("columns", "values", "units", "extract")) %>%
    gt::fmt_missing(columns = "status_color", missing_text = "") %>%
    gt::cols_hide(columns = c("W_val", "S_val", "N_val", "active", "eval")) %>%
    gt::text_transform(
      locations = gt::cells_body(columns = "units"),
      fn = function(x) {
        dplyr::case_when(
          x == "&mdash;" ~ x,
          TRUE ~ paste0("<code>", x, "</code>")
        )
      }
    ) %>%
    gt::tab_style(
      style = gt::cell_text(
        size = gt::px(28),
        weight = 500,
        align = "left",
        color = "#444444"
      ),
      locations = gt::cells_title("title")
    ) %>%
    gt::tab_style(
      style = gt::cell_text(
        size = gt::px(12),
        align = "left"
      ),
      locations = gt::cells_title("subtitle")
    ) %>%
    gt::tab_style(
      style = gt::cell_text(
        weight = "bold",
        color = "#666666",
        size = ifelse(size == "small", gt::px(10), gt::px(13))
      ),
      locations = gt::cells_body(columns = "i")
    ) %>%
    gt::tab_style(
      style = gt::cell_fill(color = "#4CA64C"),
      locations = gt::cells_body(
        columns = "status_color",
        rows = units == n_pass
      )
    ) %>%
    gt::tab_style(
      style = gt::cell_fill(color = "#4CA64C66", alpha = 0.5),
      locations = gt::cells_body(
        columns = "status_color",
        rows = units != n_pass
      )
    ) %>%
    gt::tab_style(
      style = gt::cell_fill(color = "#FFBF00"),
      locations = gt::cells_body(
        columns = "status_color",
        rows = W_val
      )
    ) %>%
    gt::tab_style(
      style = gt::cell_fill(color = "#CF142B"),
      locations = gt::cells_body(
        columns = "status_color",
        rows = S_val
      )
    ) %>%
    gt::tab_style(
      style = list(
        gt::cell_borders(sides = "left", color = "#D3D3D3"),
        gt::cell_fill(color = "#FCFCFC")
      ),
      locations = gt::cells_body(columns = c("precon", "W"))
    ) %>%
    gt::tab_style(
      style = gt::cell_borders(
        sides = "left",
        color = "#E5E5E5",
        style = "dashed"
      ),
      locations = gt::cells_body(columns = c("n_pass", "n_fail"))
    ) %>%
    gt::tab_style(
      style = list(
        gt::cell_borders(sides = "right", color = "#D3D3D3"),
        gt::cell_fill(color = "#FCFCFC")
      ),
      locations = gt::cells_body(columns = c("eval_sym", "N"))
    ) %>%
    gt::tab_style(
      style = gt::cell_fill(color = "#FCFCFC"),
      locations = gt::cells_body(columns = "S")
    ) %>%
    gt::tab_style(
      style = gt::cell_borders(
        sides = "left",
        color = "#E5E5E5",
        style = "dashed"
      ),
      locations = list(
        gt::cells_body(columns = c("columns", "values"))
      )
    ) %>%
    gt::tab_style(
      style = list(
        gt::cell_fill(color = "#F2F2F2", alpha = 0.75),
        gt::cell_text(color = "#8B8B8B")
      ),
      locations = gt::cells_body(
        columns = gt::everything(),
        rows = active == FALSE
      )
    ) %>%
    gt::tab_style(
      style = gt::cell_fill(color = "#FFC1C1", alpha = 0.35),
      locations = gt::cells_body(
        columns = gt::everything(),
        rows = eval == "ERROR"
      )
    ) %>%
    gt::tab_style(
      style = gt::cell_text(size = gt::px(11)),
      locations = gt::cells_body(
        columns = c("units", "n_pass", "n_fail")
      )
    )
  
  if (!has_agent_intel(agent)) {
    
    agent_report <-
      agent_report %>%
      gt::text_transform(
        locations = gt::cells_body(
          columns = c(
            "precon", "eval_sym", "units", "f_pass", "f_fail",
            "n_pass", "n_fail", "W", "S", "N", "extract"
          )
        ),
        fn = function(x) {
          ""
        }
      ) %>%
      gt::tab_style(
        style = list(
          gt::cell_fill(color = "#F2F2F2"),
          gt::cell_borders(
            sides = "right",
            style = "solid",
            color = "#F2F2F2"
          )
        ),
        locations = gt::cells_body(
          columns = c(
            "precon", "eval_sym", "units", "f_pass", "f_fail",
            "n_pass", "n_fail", "W", "S", "N", "extract"
          )
        )
      ) %>%
      gt::tab_header(
        title = gt::md(
          paste0(
            "<div>",
            "<span style=\"float: left;\">", 
            get_lsv("agent_report/pointblank_validation_plan_text")[[lang]],
            "</span>",
            "<span style=\"float: right; text-decoration-line: underline; ",
            "text-underline-position: under;",
            "font-size: 16px; text-decoration-color: #9C2E83;",
            "padding-top: 0.1em; padding-right: 0.4em;\">",
            get_lsv("agent_report/no_interrogation_performed_text")[[lang]],
            "</span>",
            "</div>"
          )
        ),
        subtitle = gt::md(
          paste0(
            agent_label_styled, " ", table_type, " ", table_time, " <br><br>"
          )
        )
      )
  }
  
  if (size == "small") {
    
    agent_report <- 
      agent_report %>%
      gt::cols_hide(c("columns", "values", "eval_sym", "precon", "extract")) %>%
      gt::cols_width(
        "status_color" ~ gt::px(4),
        "i" ~ gt::px(25),
        "type" ~ gt::px(190),
        "precon" ~ gt::px(30),
        "units" ~ gt::px(50),
        "n_pass" ~ gt::px(50),
        "n_fail" ~ gt::px(50),
        "W" ~ gt::px(30),
        "S" ~ gt::px(30),
        "N" ~ gt::px(30),
        gt::everything() ~ gt::px(20)
      ) %>%
      gt::tab_style(
        locations = gt::cells_body(columns = gt::everything()),
        style = "height: 35px"
      )
    
    if (!has_agent_intel(agent)) {
      
      agent_report <- 
        agent_report %>%
        gt::tab_header(
          title = gt::md(
            paste0(
              "<div>",
              "<span style=\"float: left;\">", 
              get_lsv("agent_report/pointblank_validation_plan_text")[[lang]],
              "</span>",
              "<span style=\"float: right; text-decoration-line: underline; ",
              "font-size: 16px; text-decoration-color: #008B8B;",
              "padding-top: 0.1em; padding-right: 0.4em;\">",
              get_lsv("agent_report/no_interrogation_performed_text")[[lang]],
              "</span>",
              "</div>"
            )
          ),
          subtitle = gt::md(
            paste0(agent_label_styled, " ", table_type, " <br><br>")
          )
        )
      
    } else {
      
      agent_report <- 
        agent_report %>%
        gt::tab_header(
          title = get_lsv(text = c(
            "agent_report", "pointblank_validation_title_text"
          ))[[lang]],
          subtitle = gt::md(
            paste0(agent_label_styled, " ", table_type, " <br><br>")
          )
        )
    }
    
  } else {
    
    agent_report <- 
      agent_report %>%
      gt::cols_width(
        "status_color" ~ gt::px(6),
        "i" ~ gt::px(35),
        "type" ~ gt::px(190),
        "columns" ~ gt::px(120),
        "values" ~ gt::px(120),
        "precon" ~ gt::px(50),
        "eval_sym" ~ gt::px(50),
        "W" ~ gt::px(30),
        "S" ~ gt::px(30),
        "N" ~ gt::px(30),
        "extract" ~ gt::px(65),
        gt::everything() ~ gt::px(50)
      ) %>%
      gt::tab_style(
        style = gt::cell_text(weight = "bold", color = "#666666"),
        locations = gt::cells_column_labels(columns = gt::everything())
      ) %>%
      gt::tab_style(
        locations = gt::cells_body(columns = gt::everything()),
        style = "height: 40px"
      ) %>%
      gt::opt_table_font(font = gt::google_font("IBM Plex Sans")) %>%
      gt::opt_css(
        paste0(
          "@import url(\"https://unpkg.com/",
          "balloon-css/balloon.min.css\");"
        )
      ) %>%
      gt::opt_css(
        css = "
          #pb_agent {
            -webkit-font-smoothing: antialiased;
          }
          #pb_agent .gt_row {
            overflow: visible;
          }
          #pb_agent .gt_sourcenote {
            height: 35px;
            padding: 0
          }
          #pb_agent code {
            font-family: 'IBM Plex Mono', monospace, courier;
            background-color: transparent;
            padding: 0;
          }
        "
      )
  }
  
  class(agent_report) <- c("ptblank_agent_report", class(agent_report))
  
  # nocov end
  
  agent_report
}

get_default_title_text <- function(report_type,
                                   lang) {
  
  if (report_type == "informant") {
    title_text <- 
      get_lsv(text = c(
        "informant_report",
        "pointblank_information_title_text"
      ))[[lang]]
    
  } else if (report_type == "agent") {
    
    title_text <- 
      get_lsv(text = c(
        "agent_report",
        "pointblank_validation_title_text"
      ))[[lang]]
    
  } else if (grepl("multiagent", report_type)) {
    
    title_text <- 
      get_lsv(text = c(
        "multiagent_report",
        "pointblank_multiagent_title_text"
      ))[[lang]]
  }
  
  title_text
}

process_title_text <- function(title,
                               tbl_name,
                               report_type,
                               lang) {
  
  if (report_type == "multiagent:wide") {
    if (title == ":tbl_name:") {
      stop(
        "The `:tbl_name:` option can't be used with `get_multiagent_report()`.",
        call. = FALSE
      )
    }
  }
  
  if (is.null(title)) {
    title_text <- ""
  } else if (is.na(title)) {
    title_text <- ""
  } else if (title == ":default:") {
    title_text <- get_default_title_text(report_type = report_type, lang = lang)
  } else if (title == ":none:") {
    title_text <- ""
  } else if (title == ":tbl_name:") {
    if (!is.na(tbl_name) && tbl_name != "NA") {
      title_text <- gt::md(paste0("<code>", tbl_name, "</code>"))
    } else {
      title_text <- ""
    }
  } else if (inherits(title, "AsIs")) {
    title_text <- unclass(title)
  } else if (inherits(title, "character")) {
    title_text <- gt::md(title)
  }
  title_text
}

create_table_time_html <- function(time_start,
                                   time_end,
                                   size = "standard",
                                   locale = NULL) {
  
  if (length(time_start) < 1) {
    return("")
  }
  
  time_duration <- get_time_duration(time_start, time_end)
  time_duration_formatted <- 
    print_time_duration_report(time_duration, locale = locale)
  
  as.character(
    htmltools::tagList(
      htmltools::tags$span(
        style = htmltools::css(
          `background-color` = "#FFF",
          color = "#444",
          padding = if (size == "standard") "0.5em 0.5em" else "",
          position = "inherit",
          `text-transform` = "uppercase",
          `margin-left` = if (size == "standard") "10px" else "",
          border = if (size == "standard") "solid 1px #999999" else "",
          `font-variant-numeric` = "tabular-nums",
          `border-radius` = "0",
          padding = "2px 10px 2px 10px",
          padding = if (size == "standard") {
            "2px 10px 2px 10px" 
          } else {
            "2px 10px 2px 5px"
          },
          `border-right` = if (size == "small") "solid 1px #333" else ""
        ),
        format(time_start, "%Y-%m-%d %H:%M:%S %Z")
      ),
      htmltools::tags$span(
        style = htmltools::css(
          `background-color` = "#FFF",
          color = "#444",
          padding = if (size == "standard") "0.5em 0.5em" else "",
          position = "inherit",
          margin = "5px 1px 5px 0",
          border = if (size == "standard") "solid 1px #999999" else "",
          `border-left` = if (size == "small") "none" else "",
          `font-variant-numeric` = "tabular-nums",
          `border-radius` = "0",
          padding = "2px 10px 2px 10px"
        ),
        time_duration_formatted,
      ),
      htmltools::tags$span(
        style = htmltools::css(
          `background-color` = "#FFF",
          color = "#444",
          padding = if (size == "standard") "0.5em 0.5em" else "",
          position = "inherit",
          `text-transform` = "uppercase",
          `margin` = "5px 1px 5px -1px",
          border = if (size == "standard") "solid 1px #999999" else "",
          `font-variant-numeric` = "tabular-nums",
          `border-left` = if (size == "small") "solid 1px #333" else "",
          `border-radius` = "0",
          padding = "2px 10px 2px 10px"
        ),
        format(time_end, "%Y-%m-%d %H:%M:%S %Z")
      )
    )
  )
}

print_time_duration_report <- function(time_diff_s,
                                       locale = NULL) {
  
  if (time_diff_s < 1) {
    "< 1 s"
  } else {
    paste0(
      pb_fmt_number(
        round(time_diff_s, 1),
        decimals = 1, locale = locale
      ),
      " s"
    )
  }
}

create_agent_label_html <- function(agent) {
  
  htmltools::tags$span(
    agent$label,
    style = htmltools::css(
      `text-decoration-style` = "solid",
      `text-decoration-color` = "#ADD8E6",
      `text-decoration-line` = "underline",
      `text-underline-position` = "under",
      color = "#333333",
      `font-variant-numeric` = "tabular-nums",
      `padding-left` = "4px",
      `margin-right` = "5px",
      `padding-right` = "2px"
    )
  ) %>% as.character()
}

create_table_type_html <- function(tbl_src, tbl_name) {
  
  text <- 
    switch(
      tbl_src,
      data.frame = c("#9933CC", "#FFFFFF", "data frame"),
      tbl_df = c("#F1D35A", "#222222", "tibble"),
      sqlite = c("#BACBEF", "#222222", "SQLite"),
      duckdb = c("#000000", "#FFFFFF", "DuckDB"),
      mysql = c("#EBAD40", "#222222", "MySQL"),
      postgres = c("#3E638B", "#FFFFFF", "PostgreSQL"),
      tbl_spark = c("#E66F21", "#FFFFFF", "Spark DataFrame"),
      Arrow = c("#353A3F", "#FFFFFF", "Apache Arrow"),
      c("#E2E2E2", "#222222", tbl_src)
    )

  if (all(!is.na(text)) && (is.na(tbl_name) || tbl_name == "NA")) {
    
    paste0(
      "<span style=\"background-color: ", text[1], ";",
      "color: ", text[2], ";padding: 0.5em 0.5em;",
      "position: inherit;text-transform: uppercase;margin: 5px 1px 5px 4px;",
      "font-weight: bold;border: solid 1px ", text[1], ";",
      "padding: 2px 10px 2px 10px;font-size: smaller;\">",
      text[3],
      "</span>"
    )
  } else if (all(!is.na(text)) && !is.na(tbl_name)) {
    
    as.character(
      htmltools::tagList(
        htmltools::tags$span(
          text[3],
          style = htmltools::css(
            `background-color` = text[1],
            color = text[2],
            padding = "0.5em 0.5em",
            position = "inherit",
            `text-transform` = "uppercase",
            margin = "5px 0px 5px 5px",
            `font-weight` = "bold",
            border = paste0("solid 1px ", text[1]),
            padding = "2px 15px 2px 15px",
            `font-size` = "smaller"
          )
        ),
        htmltools::tags$span(
          tbl_name,
          style = htmltools::css(
            `background-color` = "none",
            color = "#222222",
            padding = "0.5em 0.5em",
            position = "inherit",
            margin = "5px 10px 5px -4px",
            `font-weight` = "bold",
            border = paste0("solid 1px ", text[1]),
            padding = "2px 15px 2px 15px",
            `font-size` = "smaller"
          )
        )
      )
    )
    
  } else {
    ""
  }
}

make_action_levels_html <- function(agent, locale) {
  
  actions <- agent$actions
  
  if (is.null(unlist(actions[1:6]))) {
    return("")
  }

  warn <- 
    c(
      pb_fmt_number(actions$warn_fraction, decimals = 2, locale = locale),
      pb_fmt_number(actions$warn_count, decimals = 0, locale = locale)
    ) %||% "&mdash;"
  
  stop <-
    c(
      pb_fmt_number(actions$stop_fraction, decimals = 2, locale = locale),
      pb_fmt_number(actions$stop_count, decimals = 0, locale = locale)
    ) %||% "&mdash;"
  
  
  notify <-
    c(
      pb_fmt_number(actions$notify_fraction, decimals = 2, locale = locale),
      pb_fmt_number(actions$notify_count, decimals = 0, locale = locale)
    ) %||% "&mdash;"
  
  as.character(
    htmltools::tagList(
      htmltools::tags$span(
        "WARN",
        style = htmltools::css(
          `background-color` = "#E5AB00",
          color = "white",
          padding = "0.5em 0.5em",
          position = "inherit",
          `text-transform` = "uppercase",
          margin = "5px 0px 5px 5px",
          `font-weight` = "bold",
          border = paste0("solid 1px #E5AB00"),
          padding = "2px 15px 2px 15px",
          `font-size` = "smaller"
        )
      ),
      htmltools::tags$span(
        htmltools::HTML(warn),
        style = htmltools::css(
          `background-color` = "none",
          color = "#333333",
          padding = "0.5em 0.5em",
          position = "inherit",
          margin = "5px 0px 5px -4px",
          `font-weight` = "bold",
          border = paste0("solid 1px #E5AB00"),
          padding = "2px 15px 2px 15px",
          `font-size` = "smaller"
        )
      ),
      htmltools::tags$span(
        "STOP",
        style = htmltools::css(
          `background-color` = "#D0182F",
          color = "white",
          padding = "0.5em 0.5em",
          position = "inherit",
          `text-transform` = "uppercase",
          margin = "5px 0px 5px 1px",
          `font-weight` = "bold",
          border = paste0("solid 1px #D0182F"),
          padding = "2px 15px 2px 15px",
          `font-size` = "smaller"
        )
      ),
      htmltools::tags$span(
        htmltools::HTML(stop),
        style = htmltools::css(
          `background-color` = "none",
          color = "#333333",
          padding = "0.5em 0.5em",
          position = "inherit",
          margin = "5px 0px 5px -4px",
          `font-weight` = "bold",
          border = paste0("solid 1px #D0182F"),
          padding = "2px 15px 2px 15px",
          `font-size` = "smaller"
        )
      ),
      htmltools::tags$span(
        "NOTIFY",
        style = htmltools::css(
          `background-color` = "#499FFE",
          color = "white",
          padding = "0.5em 0.5em",
          position = "inherit",
          `text-transform` = "uppercase",
          margin = "5px 0px 5px 1px",
          `font-weight` = "bold",
          border = paste0("solid 1px #499FFE"),
          padding = "2px 15px 2px 15px",
          `font-size` = "smaller"
        )
      ),
      htmltools::tags$span(
        htmltools::HTML(notify),
        style = htmltools::css(
          `background-color` = "none",
          color = "#333333",
          padding = "0.5em 0.5em",
          position = "inherit",
          margin = "5px 0px 5px -4px",
          `font-weight` = "bold",
          border = paste0("solid 1px #499FFE"),
          padding = "2px 15px 2px 15px",
          `font-size` = "smaller"
        )
      )
    )
  )
}

make_boxed_text_html <- function(x,
                                 size = "standard",
                                 color = "#333333",
                                 background = "transparent",
                                 font_size = "15px",
                                 padding = "5px",
                                 tt_text = NULL,
                                 tt_position = "left",
                                 tt_text_size = NULL,
                                 border_radius = NULL,
                                 v_align = "middle") {
  
  if (!is.null(tt_position) && size == "standard") {
    text_type <- "aria-label"
  } else {
    text_type <- "title"
  }
  
  text_html <- 
    htmltools::tags$span(
      htmltools::HTML(x),
      style = htmltools::css(
        background = background,
        padding = padding,
        color = color,
        `vertical-align` = v_align,
        `font-size` = font_size,
        border = "none",
        `border-radius` = border_radius
      )
    )
  
  if (size == "standard") {
    text_html <-
      text_html %>%
      htmltools::tagAppendAttributes(
        `aria-label` = tt_text,
        `data-balloon-pos` = tt_position,
        `data-balloon-length` = if (!is.null(tt_text_size)) tt_text_size
      )
  } else {
    text_html <- text_html %>% htmltools::tagAppendAttributes(`title` = tt_text)
  }
  
  text_html %>% as.character()
}

icon_status <- function(icon = c("unchanged", "modified", "segmented")) {
  
  icon <- match.arg(icon)
  
  htmltools::HTML(
    paste(
      readLines(
        con = system.file(
          "img", "status_icons", paste0(icon, ".svg"),
          package = "pointblank"
        ), 
        warn = FALSE
      ), collapse = ""
    )
  ) %>%
    as.character()
}
